# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: opencensus/proto/metrics/v1/metrics.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2
from . import resource_pb2 as opencensus_dot_proto_dot_resource_dot_v1_dot_resource__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='opencensus/proto/metrics/v1/metrics.proto',
  package='opencensus.proto.metrics.v1',
  syntax='proto3',
  serialized_options=_b('\n\036io.opencensus.proto.metrics.v1B\014MetricsProtoP\001ZDgithub.com/census-instrumentation/opencensus-proto/gen-go/metrics/v1\352\002\033OpenCensus.Proto.Metrics.V1'),
  serialized_pb=_b('\n)opencensus/proto/metrics/v1/metrics.proto\x12\x1bopencensus.proto.metrics.v1\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1egoogle/protobuf/wrappers.proto\x1a+opencensus/proto/resource/v1/resource.proto\"\xf1\x01\n\x06Metric\x12Z\n\x11metric_descriptor\x18\x01 \x01(\x0b\x32-.opencensus.proto.metrics.v1.MetricDescriptorR\x10metricDescriptor\x12G\n\ntimeseries\x18\x02 \x03(\x0b\x32\'.opencensus.proto.metrics.v1.TimeSeriesR\ntimeseries\x12\x42\n\x08resource\x18\x03 \x01(\x0b\x32&.opencensus.proto.resource.v1.ResourceR\x08resource\"\x96\x03\n\x10MetricDescriptor\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12 \n\x0b\x64\x65scription\x18\x02 \x01(\tR\x0b\x64\x65scription\x12\x12\n\x04unit\x18\x03 \x01(\tR\x04unit\x12\x46\n\x04type\x18\x04 \x01(\x0e\x32\x32.opencensus.proto.metrics.v1.MetricDescriptor.TypeR\x04type\x12\x44\n\nlabel_keys\x18\x05 \x03(\x0b\x32%.opencensus.proto.metrics.v1.LabelKeyR\tlabelKeys\"\xa9\x01\n\x04Type\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x0f\n\x0bGAUGE_INT64\x10\x01\x12\x10\n\x0cGAUGE_DOUBLE\x10\x02\x12\x16\n\x12GAUGE_DISTRIBUTION\x10\x03\x12\x14\n\x10\x43UMULATIVE_INT64\x10\x04\x12\x15\n\x11\x43UMULATIVE_DOUBLE\x10\x05\x12\x1b\n\x17\x43UMULATIVE_DISTRIBUTION\x10\x06\x12\x0b\n\x07SUMMARY\x10\x07\">\n\x08LabelKey\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12 \n\x0b\x64\x65scription\x18\x02 \x01(\tR\x0b\x64\x65scription\"\xd9\x01\n\nTimeSeries\x12\x43\n\x0fstart_timestamp\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x0estartTimestamp\x12J\n\x0clabel_values\x18\x02 \x03(\x0b\x32\'.opencensus.proto.metrics.v1.LabelValueR\x0blabelValues\x12:\n\x06points\x18\x03 \x03(\x0b\x32\".opencensus.proto.metrics.v1.PointR\x06points\"?\n\nLabelValue\x12\x14\n\x05value\x18\x01 \x01(\tR\x05value\x12\x1b\n\thas_value\x18\x02 \x01(\x08R\x08hasValue\"\xc5\x02\n\x05Point\x12\x38\n\ttimestamp\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\ttimestamp\x12!\n\x0bint64_value\x18\x02 \x01(\x03H\x00R\nint64Value\x12#\n\x0c\x64ouble_value\x18\x03 \x01(\x01H\x00R\x0b\x64oubleValue\x12_\n\x12\x64istribution_value\x18\x04 \x01(\x0b\x32..opencensus.proto.metrics.v1.DistributionValueH\x00R\x11\x64istributionValue\x12P\n\rsummary_value\x18\x05 \x01(\x0b\x32).opencensus.proto.metrics.v1.SummaryValueH\x00R\x0csummaryValueB\x07\n\x05value\"\xcb\x06\n\x11\x44istributionValue\x12\x14\n\x05\x63ount\x18\x01 \x01(\x03R\x05\x63ount\x12\x10\n\x03sum\x18\x02 \x01(\x01R\x03sum\x12\x37\n\x18sum_of_squared_deviation\x18\x03 \x01(\x01R\x15sumOfSquaredDeviation\x12\x63\n\x0e\x62ucket_options\x18\x04 \x01(\x0b\x32<.opencensus.proto.metrics.v1.DistributionValue.BucketOptionsR\rbucketOptions\x12O\n\x07\x62uckets\x18\x05 \x03(\x0b\x32\x35.opencensus.proto.metrics.v1.DistributionValue.BucketR\x07\x62uckets\x1a\xa0\x01\n\rBucketOptions\x12\x63\n\x08\x65xplicit\x18\x01 \x01(\x0b\x32\x45.opencensus.proto.metrics.v1.DistributionValue.BucketOptions.ExplicitH\x00R\x08\x65xplicit\x1a\"\n\x08\x45xplicit\x12\x16\n\x06\x62ounds\x18\x01 \x03(\x01R\x06\x62oundsB\x06\n\x04type\x1as\n\x06\x42ucket\x12\x14\n\x05\x63ount\x18\x01 \x01(\x03R\x05\x63ount\x12S\n\x08\x65xemplar\x18\x02 \x01(\x0b\x32\x37.opencensus.proto.metrics.v1.DistributionValue.ExemplarR\x08\x65xemplar\x1a\x86\x02\n\x08\x45xemplar\x12\x14\n\x05value\x18\x01 \x01(\x01R\x05value\x12\x38\n\ttimestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\ttimestamp\x12j\n\x0b\x61ttachments\x18\x03 \x03(\x0b\x32H.opencensus.proto.metrics.v1.DistributionValue.Exemplar.AttachmentsEntryR\x0b\x61ttachments\x1a>\n\x10\x41ttachmentsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"\xef\x03\n\x0cSummaryValue\x12\x31\n\x05\x63ount\x18\x01 \x01(\x0b\x32\x1b.google.protobuf.Int64ValueR\x05\x63ount\x12.\n\x03sum\x18\x02 \x01(\x0b\x32\x1c.google.protobuf.DoubleValueR\x03sum\x12N\n\x08snapshot\x18\x03 \x01(\x0b\x32\x32.opencensus.proto.metrics.v1.SummaryValue.SnapshotR\x08snapshot\x1a\xab\x02\n\x08Snapshot\x12\x31\n\x05\x63ount\x18\x01 \x01(\x0b\x32\x1b.google.protobuf.Int64ValueR\x05\x63ount\x12.\n\x03sum\x18\x02 \x01(\x0b\x32\x1c.google.protobuf.DoubleValueR\x03sum\x12q\n\x11percentile_values\x18\x03 \x03(\x0b\x32\x44.opencensus.proto.metrics.v1.SummaryValue.Snapshot.ValueAtPercentileR\x10percentileValues\x1aI\n\x11ValueAtPercentile\x12\x1e\n\npercentile\x18\x01 \x01(\x01R\npercentile\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05valueB\x94\x01\n\x1eio.opencensus.proto.metrics.v1B\x0cMetricsProtoP\x01ZDgithub.com/census-instrumentation/opencensus-proto/gen-go/metrics/v1\xea\x02\x1bOpenCensus.Proto.Metrics.V1b\x06proto3')
  ,
  dependencies=[google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,google_dot_protobuf_dot_wrappers__pb2.DESCRIPTOR,opencensus_dot_proto_dot_resource_dot_v1_dot_resource__pb2.DESCRIPTOR,])



_METRICDESCRIPTOR_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='opencensus.proto.metrics.v1.MetricDescriptor.Type',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GAUGE_INT64', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GAUGE_DOUBLE', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='GAUGE_DISTRIBUTION', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CUMULATIVE_INT64', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CUMULATIVE_DOUBLE', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CUMULATIVE_DISTRIBUTION', index=6, number=6,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SUMMARY', index=7, number=7,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=666,
  serialized_end=835,
)
_sym_db.RegisterEnumDescriptor(_METRICDESCRIPTOR_TYPE)


_METRIC = _descriptor.Descriptor(
  name='Metric',
  full_name='opencensus.proto.metrics.v1.Metric',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='metric_descriptor', full_name='opencensus.proto.metrics.v1.Metric.metric_descriptor', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='metricDescriptor', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='timeseries', full_name='opencensus.proto.metrics.v1.Metric.timeseries', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='timeseries', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='resource', full_name='opencensus.proto.metrics.v1.Metric.resource', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='resource', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=185,
  serialized_end=426,
)


_METRICDESCRIPTOR = _descriptor.Descriptor(
  name='MetricDescriptor',
  full_name='opencensus.proto.metrics.v1.MetricDescriptor',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='opencensus.proto.metrics.v1.MetricDescriptor.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='name', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='description', full_name='opencensus.proto.metrics.v1.MetricDescriptor.description', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='description', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='unit', full_name='opencensus.proto.metrics.v1.MetricDescriptor.unit', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='unit', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='type', full_name='opencensus.proto.metrics.v1.MetricDescriptor.type', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='type', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='label_keys', full_name='opencensus.proto.metrics.v1.MetricDescriptor.label_keys', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='labelKeys', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _METRICDESCRIPTOR_TYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=429,
  serialized_end=835,
)


_LABELKEY = _descriptor.Descriptor(
  name='LabelKey',
  full_name='opencensus.proto.metrics.v1.LabelKey',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='opencensus.proto.metrics.v1.LabelKey.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='key', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='description', full_name='opencensus.proto.metrics.v1.LabelKey.description', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='description', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=837,
  serialized_end=899,
)


_TIMESERIES = _descriptor.Descriptor(
  name='TimeSeries',
  full_name='opencensus.proto.metrics.v1.TimeSeries',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='start_timestamp', full_name='opencensus.proto.metrics.v1.TimeSeries.start_timestamp', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='startTimestamp', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='label_values', full_name='opencensus.proto.metrics.v1.TimeSeries.label_values', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='labelValues', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='points', full_name='opencensus.proto.metrics.v1.TimeSeries.points', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='points', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=902,
  serialized_end=1119,
)


_LABELVALUE = _descriptor.Descriptor(
  name='LabelValue',
  full_name='opencensus.proto.metrics.v1.LabelValue',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='value', full_name='opencensus.proto.metrics.v1.LabelValue.value', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='value', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='has_value', full_name='opencensus.proto.metrics.v1.LabelValue.has_value', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='hasValue', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1121,
  serialized_end=1184,
)


_POINT = _descriptor.Descriptor(
  name='Point',
  full_name='opencensus.proto.metrics.v1.Point',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='opencensus.proto.metrics.v1.Point.timestamp', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='timestamp', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='int64_value', full_name='opencensus.proto.metrics.v1.Point.int64_value', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='int64Value', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='double_value', full_name='opencensus.proto.metrics.v1.Point.double_value', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='doubleValue', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='distribution_value', full_name='opencensus.proto.metrics.v1.Point.distribution_value', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='distributionValue', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='summary_value', full_name='opencensus.proto.metrics.v1.Point.summary_value', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='summaryValue', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='value', full_name='opencensus.proto.metrics.v1.Point.value',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=1187,
  serialized_end=1512,
)


_DISTRIBUTIONVALUE_BUCKETOPTIONS_EXPLICIT = _descriptor.Descriptor(
  name='Explicit',
  full_name='opencensus.proto.metrics.v1.DistributionValue.BucketOptions.Explicit',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='bounds', full_name='opencensus.proto.metrics.v1.DistributionValue.BucketOptions.Explicit.bounds', index=0,
      number=1, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='bounds', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1934,
  serialized_end=1968,
)

_DISTRIBUTIONVALUE_BUCKETOPTIONS = _descriptor.Descriptor(
  name='BucketOptions',
  full_name='opencensus.proto.metrics.v1.DistributionValue.BucketOptions',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='explicit', full_name='opencensus.proto.metrics.v1.DistributionValue.BucketOptions.explicit', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='explicit', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_DISTRIBUTIONVALUE_BUCKETOPTIONS_EXPLICIT, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='type', full_name='opencensus.proto.metrics.v1.DistributionValue.BucketOptions.type',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=1816,
  serialized_end=1976,
)

_DISTRIBUTIONVALUE_BUCKET = _descriptor.Descriptor(
  name='Bucket',
  full_name='opencensus.proto.metrics.v1.DistributionValue.Bucket',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='count', full_name='opencensus.proto.metrics.v1.DistributionValue.Bucket.count', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='count', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='exemplar', full_name='opencensus.proto.metrics.v1.DistributionValue.Bucket.exemplar', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='exemplar', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1978,
  serialized_end=2093,
)

_DISTRIBUTIONVALUE_EXEMPLAR_ATTACHMENTSENTRY = _descriptor.Descriptor(
  name='AttachmentsEntry',
  full_name='opencensus.proto.metrics.v1.DistributionValue.Exemplar.AttachmentsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='opencensus.proto.metrics.v1.DistributionValue.Exemplar.AttachmentsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='key', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='opencensus.proto.metrics.v1.DistributionValue.Exemplar.AttachmentsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='value', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2296,
  serialized_end=2358,
)

_DISTRIBUTIONVALUE_EXEMPLAR = _descriptor.Descriptor(
  name='Exemplar',
  full_name='opencensus.proto.metrics.v1.DistributionValue.Exemplar',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='value', full_name='opencensus.proto.metrics.v1.DistributionValue.Exemplar.value', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='value', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='opencensus.proto.metrics.v1.DistributionValue.Exemplar.timestamp', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='timestamp', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='attachments', full_name='opencensus.proto.metrics.v1.DistributionValue.Exemplar.attachments', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='attachments', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_DISTRIBUTIONVALUE_EXEMPLAR_ATTACHMENTSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2096,
  serialized_end=2358,
)

_DISTRIBUTIONVALUE = _descriptor.Descriptor(
  name='DistributionValue',
  full_name='opencensus.proto.metrics.v1.DistributionValue',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='count', full_name='opencensus.proto.metrics.v1.DistributionValue.count', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='count', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sum', full_name='opencensus.proto.metrics.v1.DistributionValue.sum', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='sum', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sum_of_squared_deviation', full_name='opencensus.proto.metrics.v1.DistributionValue.sum_of_squared_deviation', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='sumOfSquaredDeviation', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='bucket_options', full_name='opencensus.proto.metrics.v1.DistributionValue.bucket_options', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='bucketOptions', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='buckets', full_name='opencensus.proto.metrics.v1.DistributionValue.buckets', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='buckets', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_DISTRIBUTIONVALUE_BUCKETOPTIONS, _DISTRIBUTIONVALUE_BUCKET, _DISTRIBUTIONVALUE_EXEMPLAR, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1515,
  serialized_end=2358,
)


_SUMMARYVALUE_SNAPSHOT_VALUEATPERCENTILE = _descriptor.Descriptor(
  name='ValueAtPercentile',
  full_name='opencensus.proto.metrics.v1.SummaryValue.Snapshot.ValueAtPercentile',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='percentile', full_name='opencensus.proto.metrics.v1.SummaryValue.Snapshot.ValueAtPercentile.percentile', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='percentile', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='opencensus.proto.metrics.v1.SummaryValue.Snapshot.ValueAtPercentile.value', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='value', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2783,
  serialized_end=2856,
)

_SUMMARYVALUE_SNAPSHOT = _descriptor.Descriptor(
  name='Snapshot',
  full_name='opencensus.proto.metrics.v1.SummaryValue.Snapshot',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='count', full_name='opencensus.proto.metrics.v1.SummaryValue.Snapshot.count', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='count', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sum', full_name='opencensus.proto.metrics.v1.SummaryValue.Snapshot.sum', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='sum', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='percentile_values', full_name='opencensus.proto.metrics.v1.SummaryValue.Snapshot.percentile_values', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='percentileValues', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_SUMMARYVALUE_SNAPSHOT_VALUEATPERCENTILE, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2557,
  serialized_end=2856,
)

_SUMMARYVALUE = _descriptor.Descriptor(
  name='SummaryValue',
  full_name='opencensus.proto.metrics.v1.SummaryValue',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='count', full_name='opencensus.proto.metrics.v1.SummaryValue.count', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='count', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sum', full_name='opencensus.proto.metrics.v1.SummaryValue.sum', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='sum', file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='snapshot', full_name='opencensus.proto.metrics.v1.SummaryValue.snapshot', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, json_name='snapshot', file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_SUMMARYVALUE_SNAPSHOT, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2361,
  serialized_end=2856,
)

_METRIC.fields_by_name['metric_descriptor'].message_type = _METRICDESCRIPTOR
_METRIC.fields_by_name['timeseries'].message_type = _TIMESERIES
_METRIC.fields_by_name['resource'].message_type = opencensus_dot_proto_dot_resource_dot_v1_dot_resource__pb2._RESOURCE
_METRICDESCRIPTOR.fields_by_name['type'].enum_type = _METRICDESCRIPTOR_TYPE
_METRICDESCRIPTOR.fields_by_name['label_keys'].message_type = _LABELKEY
_METRICDESCRIPTOR_TYPE.containing_type = _METRICDESCRIPTOR
_TIMESERIES.fields_by_name['start_timestamp'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_TIMESERIES.fields_by_name['label_values'].message_type = _LABELVALUE
_TIMESERIES.fields_by_name['points'].message_type = _POINT
_POINT.fields_by_name['timestamp'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_POINT.fields_by_name['distribution_value'].message_type = _DISTRIBUTIONVALUE
_POINT.fields_by_name['summary_value'].message_type = _SUMMARYVALUE
_POINT.oneofs_by_name['value'].fields.append(
  _POINT.fields_by_name['int64_value'])
_POINT.fields_by_name['int64_value'].containing_oneof = _POINT.oneofs_by_name['value']
_POINT.oneofs_by_name['value'].fields.append(
  _POINT.fields_by_name['double_value'])
_POINT.fields_by_name['double_value'].containing_oneof = _POINT.oneofs_by_name['value']
_POINT.oneofs_by_name['value'].fields.append(
  _POINT.fields_by_name['distribution_value'])
_POINT.fields_by_name['distribution_value'].containing_oneof = _POINT.oneofs_by_name['value']
_POINT.oneofs_by_name['value'].fields.append(
  _POINT.fields_by_name['summary_value'])
_POINT.fields_by_name['summary_value'].containing_oneof = _POINT.oneofs_by_name['value']
_DISTRIBUTIONVALUE_BUCKETOPTIONS_EXPLICIT.containing_type = _DISTRIBUTIONVALUE_BUCKETOPTIONS
_DISTRIBUTIONVALUE_BUCKETOPTIONS.fields_by_name['explicit'].message_type = _DISTRIBUTIONVALUE_BUCKETOPTIONS_EXPLICIT
_DISTRIBUTIONVALUE_BUCKETOPTIONS.containing_type = _DISTRIBUTIONVALUE
_DISTRIBUTIONVALUE_BUCKETOPTIONS.oneofs_by_name['type'].fields.append(
  _DISTRIBUTIONVALUE_BUCKETOPTIONS.fields_by_name['explicit'])
_DISTRIBUTIONVALUE_BUCKETOPTIONS.fields_by_name['explicit'].containing_oneof = _DISTRIBUTIONVALUE_BUCKETOPTIONS.oneofs_by_name['type']
_DISTRIBUTIONVALUE_BUCKET.fields_by_name['exemplar'].message_type = _DISTRIBUTIONVALUE_EXEMPLAR
_DISTRIBUTIONVALUE_BUCKET.containing_type = _DISTRIBUTIONVALUE
_DISTRIBUTIONVALUE_EXEMPLAR_ATTACHMENTSENTRY.containing_type = _DISTRIBUTIONVALUE_EXEMPLAR
_DISTRIBUTIONVALUE_EXEMPLAR.fields_by_name['timestamp'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_DISTRIBUTIONVALUE_EXEMPLAR.fields_by_name['attachments'].message_type = _DISTRIBUTIONVALUE_EXEMPLAR_ATTACHMENTSENTRY
_DISTRIBUTIONVALUE_EXEMPLAR.containing_type = _DISTRIBUTIONVALUE
_DISTRIBUTIONVALUE.fields_by_name['bucket_options'].message_type = _DISTRIBUTIONVALUE_BUCKETOPTIONS
_DISTRIBUTIONVALUE.fields_by_name['buckets'].message_type = _DISTRIBUTIONVALUE_BUCKET
_SUMMARYVALUE_SNAPSHOT_VALUEATPERCENTILE.containing_type = _SUMMARYVALUE_SNAPSHOT
_SUMMARYVALUE_SNAPSHOT.fields_by_name['count'].message_type = google_dot_protobuf_dot_wrappers__pb2._INT64VALUE
_SUMMARYVALUE_SNAPSHOT.fields_by_name['sum'].message_type = google_dot_protobuf_dot_wrappers__pb2._DOUBLEVALUE
_SUMMARYVALUE_SNAPSHOT.fields_by_name['percentile_values'].message_type = _SUMMARYVALUE_SNAPSHOT_VALUEATPERCENTILE
_SUMMARYVALUE_SNAPSHOT.containing_type = _SUMMARYVALUE
_SUMMARYVALUE.fields_by_name['count'].message_type = google_dot_protobuf_dot_wrappers__pb2._INT64VALUE
_SUMMARYVALUE.fields_by_name['sum'].message_type = google_dot_protobuf_dot_wrappers__pb2._DOUBLEVALUE
_SUMMARYVALUE.fields_by_name['snapshot'].message_type = _SUMMARYVALUE_SNAPSHOT
DESCRIPTOR.message_types_by_name['Metric'] = _METRIC
DESCRIPTOR.message_types_by_name['MetricDescriptor'] = _METRICDESCRIPTOR
DESCRIPTOR.message_types_by_name['LabelKey'] = _LABELKEY
DESCRIPTOR.message_types_by_name['TimeSeries'] = _TIMESERIES
DESCRIPTOR.message_types_by_name['LabelValue'] = _LABELVALUE
DESCRIPTOR.message_types_by_name['Point'] = _POINT
DESCRIPTOR.message_types_by_name['DistributionValue'] = _DISTRIBUTIONVALUE
DESCRIPTOR.message_types_by_name['SummaryValue'] = _SUMMARYVALUE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Metric = _reflection.GeneratedProtocolMessageType('Metric', (_message.Message,), {
  'DESCRIPTOR' : _METRIC,
  '__module__' : 'opencensus.proto.metrics.v1.metrics_pb2'
  # @@protoc_insertion_point(class_scope:opencensus.proto.metrics.v1.Metric)
  })
_sym_db.RegisterMessage(Metric)

MetricDescriptor = _reflection.GeneratedProtocolMessageType('MetricDescriptor', (_message.Message,), {
  'DESCRIPTOR' : _METRICDESCRIPTOR,
  '__module__' : 'opencensus.proto.metrics.v1.metrics_pb2'
  # @@protoc_insertion_point(class_scope:opencensus.proto.metrics.v1.MetricDescriptor)
  })
_sym_db.RegisterMessage(MetricDescriptor)

LabelKey = _reflection.GeneratedProtocolMessageType('LabelKey', (_message.Message,), {
  'DESCRIPTOR' : _LABELKEY,
  '__module__' : 'opencensus.proto.metrics.v1.metrics_pb2'
  # @@protoc_insertion_point(class_scope:opencensus.proto.metrics.v1.LabelKey)
  })
_sym_db.RegisterMessage(LabelKey)

TimeSeries = _reflection.GeneratedProtocolMessageType('TimeSeries', (_message.Message,), {
  'DESCRIPTOR' : _TIMESERIES,
  '__module__' : 'opencensus.proto.metrics.v1.metrics_pb2'
  # @@protoc_insertion_point(class_scope:opencensus.proto.metrics.v1.TimeSeries)
  })
_sym_db.RegisterMessage(TimeSeries)

LabelValue = _reflection.GeneratedProtocolMessageType('LabelValue', (_message.Message,), {
  'DESCRIPTOR' : _LABELVALUE,
  '__module__' : 'opencensus.proto.metrics.v1.metrics_pb2'
  # @@protoc_insertion_point(class_scope:opencensus.proto.metrics.v1.LabelValue)
  })
_sym_db.RegisterMessage(LabelValue)

Point = _reflection.GeneratedProtocolMessageType('Point', (_message.Message,), {
  'DESCRIPTOR' : _POINT,
  '__module__' : 'opencensus.proto.metrics.v1.metrics_pb2'
  # @@protoc_insertion_point(class_scope:opencensus.proto.metrics.v1.Point)
  })
_sym_db.RegisterMessage(Point)

DistributionValue = _reflection.GeneratedProtocolMessageType('DistributionValue', (_message.Message,), {

  'BucketOptions' : _reflection.GeneratedProtocolMessageType('BucketOptions', (_message.Message,), {

    'Explicit' : _reflection.GeneratedProtocolMessageType('Explicit', (_message.Message,), {
      'DESCRIPTOR' : _DISTRIBUTIONVALUE_BUCKETOPTIONS_EXPLICIT,
      '__module__' : 'opencensus.proto.metrics.v1.metrics_pb2'
      # @@protoc_insertion_point(class_scope:opencensus.proto.metrics.v1.DistributionValue.BucketOptions.Explicit)
      })
    ,
    'DESCRIPTOR' : _DISTRIBUTIONVALUE_BUCKETOPTIONS,
    '__module__' : 'opencensus.proto.metrics.v1.metrics_pb2'
    # @@protoc_insertion_point(class_scope:opencensus.proto.metrics.v1.DistributionValue.BucketOptions)
    })
  ,

  'Bucket' : _reflection.GeneratedProtocolMessageType('Bucket', (_message.Message,), {
    'DESCRIPTOR' : _DISTRIBUTIONVALUE_BUCKET,
    '__module__' : 'opencensus.proto.metrics.v1.metrics_pb2'
    # @@protoc_insertion_point(class_scope:opencensus.proto.metrics.v1.DistributionValue.Bucket)
    })
  ,

  'Exemplar' : _reflection.GeneratedProtocolMessageType('Exemplar', (_message.Message,), {

    'AttachmentsEntry' : _reflection.GeneratedProtocolMessageType('AttachmentsEntry', (_message.Message,), {
      'DESCRIPTOR' : _DISTRIBUTIONVALUE_EXEMPLAR_ATTACHMENTSENTRY,
      '__module__' : 'opencensus.proto.metrics.v1.metrics_pb2'
      # @@protoc_insertion_point(class_scope:opencensus.proto.metrics.v1.DistributionValue.Exemplar.AttachmentsEntry)
      })
    ,
    'DESCRIPTOR' : _DISTRIBUTIONVALUE_EXEMPLAR,
    '__module__' : 'opencensus.proto.metrics.v1.metrics_pb2'
    # @@protoc_insertion_point(class_scope:opencensus.proto.metrics.v1.DistributionValue.Exemplar)
    })
  ,
  'DESCRIPTOR' : _DISTRIBUTIONVALUE,
  '__module__' : 'opencensus.proto.metrics.v1.metrics_pb2'
  # @@protoc_insertion_point(class_scope:opencensus.proto.metrics.v1.DistributionValue)
  })
_sym_db.RegisterMessage(DistributionValue)
_sym_db.RegisterMessage(DistributionValue.BucketOptions)
_sym_db.RegisterMessage(DistributionValue.BucketOptions.Explicit)
_sym_db.RegisterMessage(DistributionValue.Bucket)
_sym_db.RegisterMessage(DistributionValue.Exemplar)
_sym_db.RegisterMessage(DistributionValue.Exemplar.AttachmentsEntry)

SummaryValue = _reflection.GeneratedProtocolMessageType('SummaryValue', (_message.Message,), {

  'Snapshot' : _reflection.GeneratedProtocolMessageType('Snapshot', (_message.Message,), {

    'ValueAtPercentile' : _reflection.GeneratedProtocolMessageType('ValueAtPercentile', (_message.Message,), {
      'DESCRIPTOR' : _SUMMARYVALUE_SNAPSHOT_VALUEATPERCENTILE,
      '__module__' : 'opencensus.proto.metrics.v1.metrics_pb2'
      # @@protoc_insertion_point(class_scope:opencensus.proto.metrics.v1.SummaryValue.Snapshot.ValueAtPercentile)
      })
    ,
    'DESCRIPTOR' : _SUMMARYVALUE_SNAPSHOT,
    '__module__' : 'opencensus.proto.metrics.v1.metrics_pb2'
    # @@protoc_insertion_point(class_scope:opencensus.proto.metrics.v1.SummaryValue.Snapshot)
    })
  ,
  'DESCRIPTOR' : _SUMMARYVALUE,
  '__module__' : 'opencensus.proto.metrics.v1.metrics_pb2'
  # @@protoc_insertion_point(class_scope:opencensus.proto.metrics.v1.SummaryValue)
  })
_sym_db.RegisterMessage(SummaryValue)
_sym_db.RegisterMessage(SummaryValue.Snapshot)
_sym_db.RegisterMessage(SummaryValue.Snapshot.ValueAtPercentile)


DESCRIPTOR._options = None
_DISTRIBUTIONVALUE_EXEMPLAR_ATTACHMENTSENTRY._options = None
# @@protoc_insertion_point(module_scope)
